package cc.fascinated.services;

import cc.fascinated.common.DateUtils;
import cc.fascinated.common.EnumUtils;
import cc.fascinated.common.MathUtils;
import cc.fascinated.common.PaginationBuilder;
import cc.fascinated.model.leaderboard.Leaderboard;
import cc.fascinated.model.score.DeviceInformation;
import cc.fascinated.model.score.Score;
import cc.fascinated.model.score.TotalScoresResponse;
import cc.fascinated.model.score.impl.scoresaber.ScoreSaberScore;
import cc.fascinated.model.score.impl.scoresaber.ScoreSaberScoreResponse;
import cc.fascinated.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.model.token.ScoreSaberPlayerScoreToken;
import cc.fascinated.model.token.ScoreSaberScoreToken;
import cc.fascinated.model.user.User;
import cc.fascinated.model.user.UserDTO;
import cc.fascinated.model.user.history.HistoryPoint;
import cc.fascinated.model.user.hmd.DeviceController;
import cc.fascinated.model.user.hmd.DeviceHeadset;
import cc.fascinated.platform.Platform;
import cc.fascinated.repository.ScoreRepository;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
@Service
@Log4j2(topic = "Score Service")
public class ScoreService {
    public static ScoreService INSTANCE;

    /**
     * The counter service to use.
     */
    @NonNull
    private final CounterService counterService;

    /**
     * The user service to use.
     */
    @NonNull
    private final UserService userService;

    /**
     * The ScoreSaber service to use.
     */
    @NonNull
    private final ScoreSaberService scoreSaberService;

    /**
     * The score repository to use.
     */
    @NonNull
    private final ScoreRepository scoreRepository;

    @Autowired
    public ScoreService(@NonNull CounterService counterService, @NonNull UserService userService, @NonNull ScoreSaberService scoreSaberService,
                        @NonNull ScoreRepository scoreRepository) {
        INSTANCE = this;
        this.counterService = counterService;
        this.userService = userService;
        this.scoreSaberService = scoreSaberService;
        this.scoreRepository = scoreRepository;
    }

    /**
     * Gets the top ranked scores from the platform.
     *
     * @param platform   The platform to get the scores from.
     * @param scoresOnly Whether to only get the scores.
     * @return The scores.
     */
    public PaginationBuilder.Page<ScoreSaberScoreResponse> getTopRankedScores(@NonNull Platform.Platforms platform, int pageNumber, boolean scoresOnly) {
        PaginationBuilder<ScoreSaberScoreResponse> builder = new PaginationBuilder<ScoreSaberScoreResponse>().build();
        builder.itemsPerPage(15);
        builder.totalItems(() -> this.scoreRepository.getTotalRankedScores(platform));
        builder.items((fetchItems) -> {
            List<Score> foundScores = this.scoreRepository.getTopRankedScores(platform, fetchItems.getItemsPerPage(), fetchItems.skipAmount());
            List<ScoreSaberScoreResponse> scores = new ArrayList<>();
            for (Score score : foundScores) {
                ScoreSaberScore scoreSaberScore = (ScoreSaberScore) score;
                UserDTO user = scoresOnly ? null : userService.getUser(score.getPlayerId()).getAsDTO();
                Leaderboard leaderboard = scoresOnly ? null : Leaderboard.getFromScoreSaberToken(scoreSaberService.getLeaderboard(score.getLeaderboardId()));

                scores.add(ScoreSaberScoreResponse.fromScore(scoreSaberScore, user, leaderboard));
            }
            return scores;
        });
        return builder.getPage(pageNumber);
    }

    /**
     * Gets all the ranked scores from the platform.
     *
     * @param platform The platform to get the scores from.
     * @return The scores.
     */
    public List<Score> getRankedScores(@NonNull Platform.Platforms platform) {
        return scoreRepository.getRankedScores(platform);
    }

    /**
     * Updates a scores pp value.
     *
     * @param scores The scores to update.
     */
    public void updateScores(Score... scores) {
        for (Score score : scores) {
            scoreRepository.updateScorePP(score.getId(), score.getPp());
        }
    }

    /**
     * Gets the total scores for the platform.
     *
     * @param platform The platform to get the scores from.
     * @return The total scores.
     */
    public TotalScoresResponse getTotalScores(Platform.Platforms platform) {
        return new TotalScoresResponse(
                scoreRepository.getTotalScores(platform),
                scoreRepository.getTotalRankedScores(platform)
        );
    }

    /**
     * Gets the total scores for the platform and user.
     *
     * @param platform The platform to get the scores from.
     * @param user     The user to get the scores from.
     * @return The total scores.
     */
    public int getTotalScores(@NonNull Platform.Platforms platform, @NonNull User user) {
        Integer totalPlayerScores = scoreRepository.getTotalPlayerScores(platform, user.getSteamId());
        return totalPlayerScores == null ? 0 : totalPlayerScores;
    }

    /**
     * Gets the total ranked scores for the platform and user.
     *
     * @param platform The platform to get the scores from.
     * @param user     The user to get the scores from.
     * @return The total ranked scores.
     */
    public int getTotalRankedScores(@NonNull Platform.Platforms platform, @NonNull User user) {
        Integer totalPlayerRankedScores = scoreRepository.getTotalPlayerRankedScores(platform, user.getSteamId());
        return totalPlayerRankedScores == null ? 0 : totalPlayerRankedScores;
    }

    /**
     * Gets the total unranked scores for the platform and user.
     *
     * @param platform The platform to get the scores from.
     * @param user     The user to get the scores from.
     * @return The total unranked scores.
     */
    public int getTotalUnrankedScores(@NonNull Platform.Platforms platform, @NonNull User user) {
        Integer totalPlayerUnrankedScores = scoreRepository.getTotalPlayerUnrankedScores(platform, user.getSteamId());
        return totalPlayerUnrankedScores == null ? 0 : totalPlayerUnrankedScores;
    }

    /**
     * Tracks a ScoreSaber score.
     *
     * @param token The token of the score to track.
     */
    public void trackScoreSaberScore(ScoreSaberPlayerScoreToken token) {
        ScoreSaberLeaderboardToken leaderboard = this.scoreSaberService.getLeaderboard(token.getLeaderboard().getId());
        ScoreSaberScoreToken score = token.getScore();
        User user = userService.getUser(score.getLeaderboardPlayerInfo().getId());

        double accuracy = leaderboard.getMaxScore() != 0 ? ((double) score.getBaseScore() / leaderboard.getMaxScore()) * 100 : 0;
        if (accuracy == 0) {
            log.warn("[Scoresaber] Leaderboard '{}' has a max score of 0, unable to calculate accuracy :(", leaderboard.getId());
        }
        double pp = score.getPp() != 0 ? PlatformService.INSTANCE.getScoreSaberPlatform().getPp(leaderboard.getStars(), accuracy) : 0; // Recalculate the pp
        String[] modifiers = !score.getModifiers().isEmpty() ? score.getModifiers().split(",") : new String[0];

        DeviceHeadset deviceHmd;
        boolean useLegacyHmdIdentification = score.getDeviceHmd() == null && score.getHmd() != 0;
        if (!useLegacyHmdIdentification) { // Use the new format
            deviceHmd = DeviceHeadset.getByName(score.getDeviceHmd());
        } else { // Use the legacy format (only includes the HMD, missing controller information)
            deviceHmd = DeviceHeadset.getByFallbackValue(score.getHmd());
        }

        ScoreSaberScore scoreSaberScore = new ScoreSaberScore(
                this.counterService.getNext(CounterService.CounterType.SCORE),
                user.getSteamId(),
                Platform.Platforms.SCORESABER,
                score.getId(),
                leaderboard.getId(),
                score.getRank(),
                accuracy,
                pp == 0 ? null : pp, // no pp, set to null to save data
                score.getBaseScore(),
                modifiers.length == 0 ? null : modifiers, // no modifiers, set to null to save data
                score.getMissedNotes() == 0 ? null : score.getMissedNotes(), // no misses, set to null to save data
                score.getBadCuts() == 0 ? null : score.getBadCuts(), // no bad cuts, set to null to save data
                new DeviceInformation(
                        deviceHmd,
                        score.getDeviceControllerLeft() == null ? DeviceController.UNKNOWN : DeviceController.getByName(score.getDeviceControllerLeft()),
                        score.getDeviceControllerRight() == null ? DeviceController.UNKNOWN : DeviceController.getByName(score.getDeviceControllerRight())
                ),
                DateUtils.getDateFromIsoString(score.getTimeSet()),
                score.getWeight() == 0 ? null : score.getWeight(), // no weight, set to null to save data
                score.getMultiplier(),
                score.getMaxCombo()
        );
        this.saveScore(user, scoreSaberScore);
        this.logScore(Platform.Platforms.SCORESABER, Leaderboard.getFromScoreSaberToken(leaderboard), scoreSaberScore, user);
        if (useLegacyHmdIdentification) {
            log.info(" - Using legacy device information, headset found: {} (missing controller information)", deviceHmd);
        }
    }

    /**
     * Gets the previous scores for a leaderboard.
     *
     * @param platform      The platform to get the score from.
     * @param user          The user to get the score from.
     * @param leaderboardId The leaderboard id to get the score from.
     * @return The previous score.
     */
    public @NonNull List<Score> getScoreHistory(@NonNull Platform.Platforms platform, @NonNull User
            user, @NonNull String leaderboardId) {
        List<Score> foundScores = new ArrayList<>(this.scoreRepository.findScores(platform, user.getSteamId(), leaderboardId));

        // Sort previous scores by timestamp (newest -> oldest)
        foundScores.sort(Comparator.comparing(Score::getTimestamp).reversed());
        return foundScores;
    }

    /**
     * Saves a score.
     *
     * @param score The score to save.
     */
    public void saveScore(User user, Score score) {
        HistoryPoint todayHistory = user.getTodayHistory();
        if (score.isRanked()) {
            todayHistory.incrementRankedPlayCount();
        } else {
            todayHistory.incrementUnrankedPlayCount();
        }
        Platform.Platforms platform = score.getPlatform();
        todayHistory.setTotalPlayCount(this.getTotalScores(platform, user)); // Get the total scores for the platform
        todayHistory.setTotalRankedPlayCount(this.getTotalRankedScores(platform, user)); // Get the total ranked scores for the platform
        todayHistory.setTotalUnrankedPlayCount(this.getTotalUnrankedScores(platform, user)); // Get the total unranked scores for the platform
        userService.saveUser(user); // Save the user
        scoreRepository.save(score); // Save the score
    }

    /**
     * Logs a score.
     *
     * @param platform The platform the score was tracked on.
     * @param score    The score to log.
     * @param user     The user who set the score.
     */
    private void logScore(@NonNull Platform.Platforms platform, @NonNull Leaderboard leaderboard, @NonNull Score
            score,
                          @NonNull User user) {
        String platformName = EnumUtils.getEnumName(platform);
        boolean isRanked = score.getPp() != 0;

        log.info("[{}] Tracked{} Score! id: {}, acc: {}%, {} score id: {},{} leaderboard: {}, difficulty: {}, player: {} ({})",
                platformName,
                isRanked ? " Ranked" : "",
                score.getId(),
                MathUtils.format(score.getAccuracy(), 2),
                platformName.toLowerCase(), score.getPlatformScoreId(),
                isRanked ? " pp: %s,".formatted(MathUtils.format(score.getPp(), 2)) : "",
                score.getLeaderboardId(),
                leaderboard.getDifficulty().getDifficulty(),
                user.getUsername() == null ? user.getSteamId() : user.getUsername(),
                user.getSteamId()
        );
    }
}
